#------------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# parsers.pyx
#
# Cython file defining the classes used for parsing connect strings and
# statements (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

# a set of alternative parameter names that are used in connect descriptors
# the key is the parameter name used in connect descriptors and the value is
# the key used in argument dictionaries (and stored in the parameter objects)
ALTERNATIVE_PARAM_NAMES = {
    "pool_connection_class": "cclass",
    "pool_purity": "purity",
    "server": "server_type",
    "transport_connect_timeout": "tcp_connect_timeout",
    "my_wallet_directory": "wallet_location"
}

# a set of parameter names used as containers for storing additional parameters
CONTAINER_PARAM_NAMES = set([
    "address",
    "address_list",
    "connect_data",
    "description",
    "description_list",
    "security",
])

# DESCRIPTION parameter names that are supported by the driver; all other
# key/value pairs are passed unchanged to the database
DESCRIPTION_PARAM_NAMES = set([
    "address",
    "address_list",
    "connect_data",
    "expire_time",
    "failover",
    "load_balance",
    "source_route",
    "retry_count",
    "retry_delay",
    "sdu",
    "tcp_connect_timeout",
    "use_sni",
    "security",
])

# extra DESCRIPTION parameter names that are passed through when detected in an
# easy connect string
EXTRA_DESCRIPTION_PARAM_NAMES = set([
    "enable",
    "recv_buf_size",
    "send_buf_size"
])

# CONNECT_DATA parameter names that are supported by the driver; all other
# key/value pairs are passed unchanged to the database
CONNECT_DATA_PARAM_NAMES = set([
    "cclass",
    "connection_id_prefix",
    "instance_name",
    "pool_boundary",
    "purity",
    "server_type",
    "service_name",
    "sid",
    "use_tcp_fast_open",
])

# SECURITY parameter names that are supported by the driver; all other
# key/value pairs are passed unchanged to the database
SECURITY_PARAM_NAMES = set([
    "ssl_server_cert_dn",
    "ssl_server_dn_match",
    "ssl_version",
    "wallet_location",
])

# a set of parameter names supported by the driver in EasyConnect strings that
# are common to all drivers
COMMON_PARAM_NAMES = set([
    "expire_time",
    "failover",
    "https_proxy",
    "https_proxy_port",
    "load_balance",
    "pool_boundary",
    "pool_connection_class",
    "pool_purity",
    "retry_count",
    "retry_delay",
    "sdu",
    "source_route",
    "ssl_server_cert_dn",
    "ssl_server_dn_match",
    "transport_connect_timeout",
    "use_sni",
    "wallet_location",
])

# extended parameter prefix
EXTENDED_PARAM_PREFIX = "pyo."

# a set of parameter names that can be specified in EasyConnect strings using
# the extended parameter prefix; these are specific to python-oracledb
EXTENDED_PARAM_NAMES = set([

    # ConnectParams
    "connection_id_prefix",
    "disable_oob",
    "driver_name",
    "edition",
    "events",
    "externalauth",
    "machine",
    "mode",
    "osuser",
    "program",
    "stmtcachesize",
    "terminal",
    "use_tcp_fast_open",

    # PoolParams
    "getmode",
    "homogeneous",
    "increment",
    "max",
    "max_lifetime_session",
    "max_sessions_per_shard",
    "min",
    "ping_interval",
    "ping_timeout",
    "soda_metadata_cache",
    "timeout",
    "wait_timeout",

])

# this cache stores the configurations acquired from one of the configuration
# stores
cdef dict cached_configs = {}

# add all of the common parameters to the extended parameters using the
# python-oracledb specific name
for name in COMMON_PARAM_NAMES:
    EXTENDED_PARAM_NAMES.add(ALTERNATIVE_PARAM_NAMES.get(name, name))


cdef class BaseParser:

    cdef Py_UCS4 get_current_char(self):
        """
        Returns the current character in the data being parsed. This method
        assumes that the end of the stream has not been encountered.
        """
        return cpython.PyUnicode_READ(self.data_kind, self.data, self.temp_pos)

    cdef int initialize(self, str data_to_parse) except -1:
        """
        Initializes the parser with the data to parse.
        """
        self.pos = 0
        self.temp_pos = 0
        self.data_as_str = data_to_parse
        self.num_chars = cpython.PyUnicode_GET_LENGTH(data_to_parse)
        self.data = cpython.PyUnicode_DATA(data_to_parse)
        self.data_kind = cpython.PyUnicode_KIND(data_to_parse)

    cdef int parse_keyword(self) except -1:
        """
        Parse a keyword from the data to parse.
        """
        cdef Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not cpython.Py_UNICODE_ISALPHA(ch) \
                    and not cpython.Py_UNICODE_ISDIGIT(ch) \
                    and ch != '_' and ch != '.':
                break
            self.temp_pos += 1

    cdef int parse_quoted_string(self, Py_UCS4 quote_type) except -1:
        """
        Parses a quoted string with the given quote type. All characters until
        the quote type is detected are discarded.
        """
        cdef Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            self.temp_pos += 1
            if ch == quote_type:
                self.pos = self.temp_pos
                return 0
        if quote_type == "'":
            errors._raise_err(errors.ERR_MISSING_ENDING_SINGLE_QUOTE)
        else:
            errors._raise_err(errors.ERR_MISSING_ENDING_DOUBLE_QUOTE)

    cdef int skip_spaces(self) except -1:
        """
        Skip any spaces that are present in the string.
        """
        cdef Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not cpython.Py_UNICODE_ISSPACE(ch):
                break
            self.temp_pos += 1


cdef int update_config_cache(str config_cache_key, dict config) except -1:
    """
    Updates the cache with the specified configuration.
    """
    cdef:
        double current_time, soft_expiry_time, hard_expiry_time
        uint32_t time_to_live, time_to_live_grace_period
        object setting

    # the config can include config_time_to_live; the default value is 86400
    # (24 hours); an explicit value of 0 disables caching
    setting = config.get("config_time_to_live")
    if setting is None:
        time_to_live = 86400
    else:
        time_to_live = int(setting)
    if time_to_live == 0:
        return 0

    # the config settings can also include config_time_to_live_grace_period;
    # the default value is 1800 (30 minutes)
    setting = config.get("config_time_to_live_grace_period")
    if setting is None:
        time_to_live_grace_period = 1800
    else:
        time_to_live_grace_period = int(setting)

    # calculate soft and hard expiry times and keep them with the config
    current_time = time.monotonic()
    soft_expiry_time = current_time + time_to_live
    hard_expiry_time = soft_expiry_time + time_to_live_grace_period
    config = copy.deepcopy(config)
    config["config_cache_soft_expiry_time"] = soft_expiry_time
    config["config_cache_hard_expiry_time"] = hard_expiry_time
    cached_configs[config_cache_key] = config


cdef class ConnectStringParser(BaseParser):

    cdef:
        DescriptionList description_list
        Description template_description
        ConnectParamsImpl params_impl
        Address template_address
        Description description
        dict parameters

    cdef int _call_protocol_hook(self, str protocol, str arg,
                                 object fn) except -1:
        """
        Check if the config cache has an entry; if an extry exists and it has
        not expired, use it; otherwise, call the protocol hook function.
        """
        cdef:
            double current_time = 0, expiry_time = 0
            dict config

        # check to see if the cache has a value and that it has not reached the
        # soft expiry time
        config = cached_configs.get(self.data_as_str)
        if config is not None:
            current_time = time.monotonic()
            expiry_time = config["config_cache_soft_expiry_time"]
            if current_time <= expiry_time:
                self.params_impl.set_from_config(config, update_cache=False)
                return 0

        # call the protocol hook function; the cache key is set on the
        # parameters instance so that calls by the hook function to
        # set_from_config() will update the cache
        params = self.params_impl._get_public_instance()
        self.params_impl._config_cache_key = self.data_as_str
        try:
            fn(protocol, arg, params)
        except Exception as e:

            # if the hook fails but a config exists in the cache and the hard
            # expiry time has not been reached the existing config is used
            if config is not None:
                expiry_time = config["config_cache_hard_expiry_time"]
                if current_time <= expiry_time:
                    self.params_impl.set_from_config(config, update_cache=False)
                    return 0
                del cached_configs[self.params_impl._config_cache_key]
            errors._raise_err(errors.ERR_PROTOCOL_HANDLER_FAILED,
                              protocol=protocol, arg=arg, cause=e)
        finally:
            self.params_impl._config_cache_key = None

    cdef bint _is_host_or_service_name_char(self, Py_UCS4 ch):
        """
        Returns whether or not the given character is allowed to be used inside
        a host name or service name.
        """
        return cpython.Py_UNICODE_ISALPHA(ch) \
                or cpython.Py_UNICODE_ISDIGIT(ch) \
                or ch in ('-', '_', '.')

    cdef int _parse_descriptor(self) except -1:
        """
        Parses a connect descriptor.
        """
        cdef:
            AddressList address_list
            Description description
            Address address
            dict args = {}
        self._parse_descriptor_key_value_pair(args)
        self.description_list = DescriptionList()
        list_args = args.get("description_list")
        if list_args is not None:
            self.description_list.set_from_args(list_args)
        else:
            list_args = args
        descriptions = list_args.get("description", list_args)
        if not isinstance(descriptions, list):
            descriptions = [descriptions]
        for desc_args in descriptions:
            description = self.template_description.copy()
            description.set_from_description_args(desc_args)
            self.description_list.children.append(description)
            sub_args = desc_args.get("connect_data")
            if sub_args is not None:
                description.set_from_connect_data_args(sub_args)
            sub_args = desc_args.get("security")
            if sub_args is not None:
                description.set_from_security_args(sub_args)
            address_lists = desc_args.get("address_list", desc_args)
            if not isinstance(address_lists, list):
                address_lists = [address_lists]
            for list_args in address_lists:
                address_list = AddressList()
                address_list.set_from_args(list_args)
                description.children.append(address_list)
                addresses = list_args.get("address", [])
                if not isinstance(addresses, list):
                    addresses = [addresses]
                for addr_args in addresses:
                    address = self.template_address.copy()
                    address.set_from_args(addr_args)
                    address_list.children.append(address)
        if not self.description_list.get_addresses():
            errors._raise_err(errors.ERR_MISSING_ADDRESS,
                              connect_string=self.data_as_str)

    cdef int _parse_descriptor_key_value_pair(self, dict args) except -1:
        """
        Parses a key-value pair from the connect string. At this point it is
        assumed that the character previously read was an opening parenthesis.
        """
        cdef:
            bint is_simple_value = False
            object value = None
            ssize_t start_pos
            Py_UCS4 ch = 0
            str name

        # parse keyword
        self.skip_spaces()
        start_pos = self.temp_pos
        self.parse_keyword()
        if self.temp_pos == start_pos:
            errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                              data=self.data_as_str)
        name = self.data_as_str[start_pos:self.temp_pos].lower()
        name = ALTERNATIVE_PARAM_NAMES.get(name, name)

        # look for equals sign
        self.skip_spaces()
        if self.temp_pos < self.num_chars:
            ch = self.get_current_char()
        if ch != '=':
            errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                              data=self.data_as_str)
        self.temp_pos += 1
        self.skip_spaces()

        # parse value
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch == '"':
                if is_simple_value:
                    errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                                      data=self.data_as_str)
                self.temp_pos += 1
                start_pos = self.temp_pos
                self.parse_quoted_string(ch)
                if self.temp_pos > start_pos + 1:
                    value = self.data_as_str[start_pos:self.temp_pos - 1]
                break
            elif ch == '(':
                if is_simple_value:
                    errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                                      data=self.data_as_str)
                self.temp_pos += 1
                if value is None:
                    value = {}
                self._parse_descriptor_key_value_pair(value)
                continue
            elif ch == ')':
                break
            elif not is_simple_value and not cpython.Py_UNICODE_ISSPACE(ch):
                if value is not None or name in CONTAINER_PARAM_NAMES:
                    errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                                      data=self.data_as_str)
                start_pos = self.temp_pos
                is_simple_value = True
            self.temp_pos += 1
        if is_simple_value:
            value = self.data_as_str[start_pos:self.temp_pos].strip()
        self.skip_spaces()
        if self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch != ')':
                errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                                  data=self.data_as_str)
            self.temp_pos += 1
        self.skip_spaces()
        self.pos = self.temp_pos

        # store value in dictionary
        if value is not None:
            self._set_descriptor_arg(args, name, value)

    cdef int _parse_easy_connect(self) except -1:
        """
        Parses an easy connect string.
        """
        cdef:
            object fn
            str protocol, arg
        protocol = self._parse_easy_connect_protocol()
        if protocol is not None:
            fn = REGISTERED_PROTOCOLS.get(protocol)
            if fn is not None:
                arg = self.data_as_str[self.temp_pos:]
                self._call_protocol_hook(protocol, arg, fn)
                self.description_list = self.params_impl.description_list
                self.pos = self.num_chars
                return 0
            elif protocol != self.template_address.protocol:
                self.template_address = self.template_address.copy()
                self.template_address.set_protocol(protocol)
        self._parse_easy_connect_hosts()
        self._parse_easy_connect_service_name()
        self._parse_easy_connect_instance_name()
        if self.description_list is not None:
            self._parse_easy_connect_parameters()

    cdef int _parse_easy_connect_host(self, Address address) except -1:
        """
        Parses a host name from the easy connect string.
        """
        cdef:
            bint found_bracket = False, found_host = False
            ssize_t start_pos = self.temp_pos
            Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not found_bracket and not found_host and ch == '[':
                found_bracket = True
                start_pos = self.temp_pos + 1
            elif found_bracket and ch == ']':
                address.host = self.data_as_str[start_pos:self.temp_pos]
                self.temp_pos = self.temp_pos + 1
                self.pos = self.temp_pos
                break
            elif found_bracket or self._is_host_or_service_name_char(ch):
                self.temp_pos += 1
                found_host = True
            else:
                if found_host:
                    address.host = self.data_as_str[start_pos:self.temp_pos]
                    self.pos = self.temp_pos
                break

    cdef int _parse_easy_connect_hosts(self) except -1:
        """
        Parses the list of hosts from an easy connect string. This should be a
        series of host names separated by commas or semicolons.
        """
        cdef:
            Address temp_address, address = None
            ssize_t i, port_index = 0
            AddressList address_list
            Py_UCS4 ch
        self.description = self.template_description.copy()
        address_list = AddressList()
        self.description.children.append(address_list)
        self.temp_pos = self.pos
        while True:
            address = self.template_address.copy()
            self._parse_easy_connect_host(address)
            if self.temp_pos != self.pos or self.pos >= self.num_chars:
                break
            self.pos = self.temp_pos
            address_list.children.append(address)
            ch = self.get_current_char()
            if ch == ':':
                self.temp_pos += 1
                self._parse_easy_connect_port(address)
                self.pos = self.temp_pos
                if self.pos >= self.num_chars:
                    break
                for i in range(port_index, len(address_list.children) - 1):
                    temp_address = address_list.children[i]
                    temp_address.port = address.port
                port_index = len(address_list.children)
                ch = self.get_current_char()
            if ch == ';':
                address_list = AddressList()
                self.description.children.append(address_list)
                port_index = 0
            elif ch != ',':
                break
            self.temp_pos += 1

    cdef int _parse_easy_connect_parameter(self) except -1:
        """
        Parses a single parameter from the easy connect string. This is
        expected to be a keyword followed by a value seprated by an equals
        sign.
        """
        cdef:
            ssize_t start_pos, end_pos = 0
            str name, value
            Py_UCS4 ch = 0
            bint keep

        # get parameter name
        self.skip_spaces()
        start_pos = self.temp_pos
        self.parse_keyword()
        if self.temp_pos == start_pos or self.temp_pos >= self.num_chars:
            return 0
        name = self.data_as_str[start_pos:self.temp_pos].lower()
        if name.startswith(EXTENDED_PARAM_PREFIX):
            name = name[len(EXTENDED_PARAM_PREFIX):]
            keep = name in EXTENDED_PARAM_NAMES
        else:
            keep = name in COMMON_PARAM_NAMES \
                    or name in EXTRA_DESCRIPTION_PARAM_NAMES
            name = ALTERNATIVE_PARAM_NAMES.get(name, name)

        # look for the equals sign
        self.skip_spaces()
        if self.temp_pos >= self.num_chars:
            return 0
        ch = self.get_current_char()
        if ch != '=':
            return 0
        self.temp_pos += 1

        # get the parameter value
        found_value = False
        self.skip_spaces()
        start_pos = self.temp_pos
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch == '"':
                # if the quote is not the first character in the value, this is
                # not a valid easy connect parameter
                if self.temp_pos > start_pos:
                    return 0
                self.temp_pos += 1
                start_pos = self.temp_pos
                self.parse_quoted_string(ch)
                end_pos = self.temp_pos - 1
                break
            elif ch == '&':
                end_pos = self.temp_pos
                break
            self.temp_pos += 1
            end_pos = self.temp_pos
        if end_pos > start_pos and keep:
            if self.parameters is None:
                self.parameters = {}
            value = self.data_as_str[start_pos:end_pos]
            if name in EXTRA_DESCRIPTION_PARAM_NAMES:
                self.parameters.setdefault("extra_args", {})[name] = value
            else:
                self.parameters[name] = value
        self.skip_spaces()
        self.pos = self.temp_pos

    cdef int _parse_easy_connect_parameters(self) except -1:
        """
        Parses the parameters from the easy connect string. This is expected to
        be a question mark followed by a series of key-value pairs separated by
        ampersands.
        """
        cdef:
            Py_UCS4 ch, expected_sep
            Address address
        expected_sep = '?'
        self.temp_pos = self.pos
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch != expected_sep:
                break
            expected_sep = '&'
            self.temp_pos += 1
            self._parse_easy_connect_parameter()

    cdef int _parse_easy_connect_port(self, Address address) except -1:
        """
        Parses a port number from the easy connect string. This consists of one
        or more digits.
        """
        cdef:
            ssize_t pos = self.temp_pos
            bint found_port = False
            Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not cpython.Py_UNICODE_ISDIGIT(ch):
                break
            found_port = True
            self.temp_pos += 1
        if found_port:
            address.port = int(self.data_as_str[pos:self.temp_pos])

    cdef str _parse_easy_connect_protocol(self):
        """
        Parses the protocol from an easy connect string. This should be a
        series of alphabetic characters or dashes, followed by a colon and two
        slashes. If such a string is found, it is saved on the template address
        associated with the parser; otherwise, the default protocol of "tcp" is
        saved on the template address associated with the parser. If no
        protocol is found, the separator (two slashes) may still be found and
        will be disarded.
        """
        cdef:
            ssize_t start_sep_pos = self.pos
            int num_sep_chars = 0
            str protocol = None
            Py_UCS4 ch
        self.temp_pos = self.pos
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch == ':':
                protocol = self.data_as_str[self.pos:self.temp_pos].lower()
                start_sep_pos = self.temp_pos + 1
            elif ch == '/' and self.temp_pos - start_sep_pos == num_sep_chars:
                num_sep_chars += 1
                if num_sep_chars == 2:
                    self.temp_pos += 1
                    self.pos = self.temp_pos
                    break
            elif not cpython.Py_UNICODE_ISALPHA(ch) and ch not in ('-', '_'):
                break
            self.temp_pos += 1
        if protocol is not None and num_sep_chars == 2:
            return protocol

    cdef str _parse_easy_connect_service_name(self):
        """
        Parses the service name from an easy connect string. This is expected
        to be a slash followed by a series of alphanumeric characters. If such
        a string is found, it is returned.
        """
        cdef:
            bint found_service_name = False, found_server_type = False
            bint found_slash = False, found_colon = False
            ssize_t service_name_end_pos = 0
            Py_UCS4 ch
            str value
        self.temp_pos = self.pos
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not found_slash and ch == '/':
                found_slash = True
            elif found_service_name and not found_colon and ch == ':':
                found_colon = True
            elif found_slash and not found_colon \
                    and self._is_host_or_service_name_char(ch):
                found_service_name = True
                service_name_end_pos = self.temp_pos + 1
            elif found_colon and cpython.Py_UNICODE_ISALPHA(ch):
                found_server_type = True
            else:
                break
            self.temp_pos += 1
        if found_service_name:
            self.description.service_name = \
                    self.data_as_str[self.pos + 1:service_name_end_pos]
        if found_slash:
            self.pos = self.temp_pos
            self.description_list = DescriptionList()
            self.description_list.children.append(self.description)
        if found_server_type:
            value = self.data_as_str[service_name_end_pos + 1:self.temp_pos]
            self.description.set_server_type(value)

    cdef str _parse_easy_connect_instance_name(self):
        """
        Parses the instance name from an easy connect string. This is expected
        to be a slash followed by a series of alphanumeric characters. If such
        a string is found, it is returned.
        """
        cdef:
            ssize_t instance_name_end_pos = 0
            bint found_instance_name = False
            bint found_slash = False
            Py_UCS4 ch
            str value
        self.temp_pos = self.pos
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if not found_slash and ch == '/':
                found_slash = True
            elif found_slash and self._is_host_or_service_name_char(ch):
                found_instance_name = True
                instance_name_end_pos = self.temp_pos + 1
            else:
                break
            self.temp_pos += 1
        if found_instance_name:
            self.description.instance_name = \
                    self.data_as_str[self.pos + 1:instance_name_end_pos]
            self.pos = self.temp_pos

    cdef dict _process_args_with_extras(self, dict args, set allowed_names,
                                        str extras_name):
        """
        Processes arguments which contain a set of known attributes and any
        number of unknown attributes. The known attributes are left untouched
        whereas the unknown ones are put in a separate dictionary with the
        given name.
        """
        cdef:
            dict extras, result = {}
            object value
            str key
        for key, value in args.items():
            if key in allowed_names:
                result[key] = value
            else:
                extras = result.setdefault(extras_name, {})
                extras[key] = value
        return result

    cdef int _set_descriptor_arg(
        self, dict args, str name, object value
    ) except -1:
        """
        Sets the arg in the dictionary. If the value is already present,
        however, a list is created and both values stored. In addition, if an
        address is being added but an address list already exists, it is simply
        added to that address list instead. Similarly, if an address list is
        being added and addresses already exist, those addresses are first
        added to the address list before the new value is added.
        """
        # process unrecognized parameters, if applicable
        if name == "description":
            value = self._process_args_with_extras(
                value, DESCRIPTION_PARAM_NAMES, "extra_args"
            )
        elif name == "connect_data":
            value = self._process_args_with_extras(
                value, CONNECT_DATA_PARAM_NAMES, "extra_connect_data_args"
            )
        elif name == "security":
            value = self._process_args_with_extras(
                value, SECURITY_PARAM_NAMES, "extra_security_args"
            )

        # add value to arguments, creating a list if encountered multiple times
        orig_value = args.get(name)
        if orig_value is None:
            if name == "address" and "address_list" in args:
                return self._set_descriptor_arg(args, "address_list",
                                                dict(address=value))
            elif name == "address_list" and "address" in args:
                addresses = args.pop("address")
                if not isinstance(addresses, list):
                    addresses = [addresses]
                value = [dict(address=a) for a in addresses] + [value]
            args[name] = value
        elif isinstance(orig_value, list):
            args[name].append(value)
        else:
            args[name] = [orig_value, value]

    cdef int parse(self, str connect_string) except -1:
        """
        Parses a connect string. If the first character is an opening
        parenthesis, the connect string is assumed to be a connect descriptor
        and any failures to parse the string as a connect descriptor will
        result in an exception. If the connect string contains key elements
        identifying it as an easy connect string, any failures to parse the
        connect string as an easy connect string will result in an exception.
        If the connect string doesn't seem to be either option, the value None
        is returned.
        """
        cdef Py_UCS4 ch
        self.initialize(connect_string)
        ch = self.get_current_char()
        if ch == '(':
            self.temp_pos += 1
            self._parse_descriptor()
        else:
            self._parse_easy_connect()
        if self.description_list is not None and self.pos != self.num_chars:
            if self.pos > 0:
                errors._raise_err(errors.ERR_CANNOT_PARSE_CONNECT_STRING,
                                  data=connect_string)
            self.description_list = None


cdef class TnsnamesFileParser(BaseParser):

    cdef int _skip_to_end_of_line(self) except -1:
        """
        Skips all characters until the next line break is found and then
        discards all whitespace after that.
        """
        cdef Py_UCS4 ch
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            self.temp_pos += 1
            if cpython.Py_UNICODE_ISLINEBREAK(ch):
                break
        self.pos = self.temp_pos
        self.skip_spaces()

    cdef str _parse_key(self):
        """
        Parses a key from the file and returns it. This consists of any number
        of non-whitespace characters until an equals sign is found. Any
        comments are discarded. If no characters are found before the equals
        sign, the line is discarded. Similarly, if an opening or closing
        parenthesis is discovered, the line is discarded.
        """
        cdef:
            bint found_key = False
            ssize_t start_pos = 0
            Py_UCS4 ch
        self.skip_spaces()
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch in ('(', ')', '#'):
                self._skip_to_end_of_line()
                found_key = False
                continue
            elif ch == '=':
                if not found_key:
                    self._skip_to_end_of_line()
                    continue
                self.temp_pos += 1
                self.pos = self.temp_pos
                return self.data_as_str[start_pos:self.temp_pos - 1].strip()
            elif not found_key:
                found_key = True
                start_pos = self.temp_pos
            self.temp_pos += 1

    cdef str _parse_value_part(self, ssize_t* num_parens):
        """
        Parses part of a value. This consists of all characters from the first
        non-whitespace character to the end of the value or the first comment.
        The number of parentheses are updated.
        """
        cdef:
            ssize_t start_pos = 0, end_pos = 0
            bint found_part = False
            Py_UCS4 ch
        self.skip_spaces()
        while self.temp_pos < self.num_chars:
            ch = self.get_current_char()
            if ch == '#':
                end_pos = self.temp_pos
                self._skip_to_end_of_line()
                if found_part:
                    break
                continue
            if found_part and num_parens[0] == 0:
                if cpython.Py_UNICODE_ISLINEBREAK(ch):
                    end_pos = self.temp_pos
                    break
            elif ch == '(':
                num_parens[0] += 1
            elif ch == ')' and num_parens[0] > 0:
                num_parens[0] -= 1
            if not found_part:
                found_part = True
                start_pos = self.temp_pos
            self.temp_pos += 1
            end_pos = self.temp_pos
        if found_part:
            return self.data_as_str[start_pos:end_pos].strip()

    cdef str _parse_value(self):
        """
        Parses a value from the file and returns it. This consists of all data
        after the first non-whitespace character until the end of the line on
        which it is found, or if the first non-whitespace character is an
        opening parenthesis, then all data until the number of opening and
        closing parentheses are equal.
        """
        cdef:
            ssize_t num_parens = 0
            list parts = []
            str part
        while self.temp_pos < self.num_chars:
            part = self._parse_value_part(&num_parens)
            if part is not None:
                parts.append(part)
            if num_parens == 0:
                break
        if parts:
            return "\n".join(parts)

    cdef int parse(self, str file_contents, object on_add_entry) except -1:
        """
        Parses the contents of a tnsnames.ora file. This consists of a series
        of key-value pairs where the keys can consist of comma-separated alias
        names and the values can be descriptors or easy connect strings. The
        method "on_add_entry" is called for each key-value pair that is
        discovered. No errors are thrown for improperly formatted files in
        order to be consistent with other drivers.
        """
        cdef str key, value
        self.initialize(file_contents)
        while self.temp_pos < self.num_chars:
            key = self._parse_key()
            value = self._parse_value()
            if key and value:
                on_add_entry(key.upper(), value.strip())



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
2600 - Module for testing timestamp variables
"""

import datetime

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            base_date = datetime.datetime(2002, 12, 9)
            date_interval = datetime.timedelta(days=i)
            date_value = base_date + date_interval
            str_value = str(i * 50)
            fsecond = int(str_value + "0" * (6 - len(str_value)))
            date_col = datetime.datetime(
                date_value.year,
                date_value.month,
                date_value.day,
                date_value.hour,
                date_value.minute,
                i * 2,
                fsecond,
            )
            if i % 2:
                date_interval = datetime.timedelta(days=i + 1)
                date_value = base_date + date_interval
                str_value = str(i * 125)
                fsecond = int(str_value + "0" * (6 - len(str_value)))
                nullable_col = datetime.datetime(
                    date_value.year,
                    date_value.month,
                    date_value.day,
                    date_value.hour,
                    date_value.minute,
                    i * 3,
                    fsecond,
                )
            else:
                nullable_col = None
            precision_col = datetime.datetime(2009, 12, 14)
            data_tuple = (i, date_col, nullable_col, precision_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_2600(self):
        "2600 - test binding in a timestamp"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            "select * from TestTimestamps where TimestampCol = :value",
            value=datetime.datetime(2002, 12, 14, 0, 0, 10, 250000),
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_2601(self):
        "2601 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            "select * from TestTimestamps where TimestampCol = :value",
            value=None,
        )
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2602(self):
        "2602 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            """
            begin
                :value := to_timestamp('20021209', 'YYYYMMDD');
            end;
            """
        )
        self.assertEqual(
            bind_vars["value"].getvalue(), datetime.datetime(2002, 12, 9)
        )

    def test_2603(self):
        "2603 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            """
            begin
                :value := :value + 5.25;
            end;
            """,
            value=datetime.datetime(2002, 12, 12, 10, 0, 0),
        )
        self.assertEqual(
            bind_vars["value"].getvalue(),
            datetime.datetime(2002, 12, 17, 16, 0, 0),
        )

    def test_2604(self):
        "2604 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            """
            begin
                :value := to_date('20021231 12:31:00',
                    'YYYYMMDD HH24:MI:SS');
            end;
            """,
            value=var,
        )
        self.assertEqual(
            var.getvalue(), datetime.datetime(2002, 12, 31, 12, 31, 0)
        )

    def test_2605(self):
        "2605 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP)
        var.setvalue(0, datetime.datetime(2002, 12, 9, 6, 0, 0))
        self.cursor.execute(
            """
            begin
                :value := :value + 5.25;
            end;
            """,
            value=var,
        )
        self.assertEqual(
            var.getvalue(), datetime.datetime(2002, 12, 14, 12, 0, 0)
        )

    def test_2606(self):
        "2606 - test cursor description is accurate"
        self.cursor.execute("select * from TestTimestamps")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            (
                "TIMESTAMPCOL",
                oracledb.DB_TYPE_TIMESTAMP,
                23,
                None,
                0,
                6,
                False,
            ),
            ("NULLABLECOL", oracledb.DB_TYPE_TIMESTAMP, 23, None, 0, 6, True),
            (
                "TIMESTAMPPRECISIONCOL",
                oracledb.DB_TYPE_TIMESTAMP,
                23,
                None,
                0,
                4,
                True,
            ),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_2607(self):
        "2607 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestTimestamps order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2608(self):
        "2608 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestTimestamps order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_2609(self):
        "2609 - test that fetching a single row returns the correct results"
        self.cursor.execute(
            """
            select *
            from TestTimestamps
            where IntCol in (3, 4)
            order by IntCol
            """
        )
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertIsNone(self.cursor.fetchone())

    def test_2610(self):
        "2610 - test binding a timestamp with zero fractional seconds"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute(
            """
            select *
            from TestTimestamps
            where trunc(TimestampCol) = :value
            """,
            value=datetime.datetime(2002, 12, 14),
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# data_types.pyx
#
# Cython file defining the messages sent to the database and the responses that
# are received by the client for establishing data type formats (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

ctypedef struct DataType:
    uint16_t data_type
    uint16_t conv_data_type
    uint16_t representation


# data types not associated with actual data in the database
cdef enum:
    TNS_DATA_TYPE_FLOAT = 4
    TNS_DATA_TYPE_STR = 5
    TNS_DATA_TYPE_VNU = 6
    TNS_DATA_TYPE_PDN = 7
    TNS_DATA_TYPE_VCS = 9
    TNS_DATA_TYPE_TIDDEF = 10
    TNS_DATA_TYPE_VBI = 15
    TNS_DATA_TYPE_UB2 = 25
    TNS_DATA_TYPE_UB4 = 26
    TNS_DATA_TYPE_SB1 = 27
    TNS_DATA_TYPE_SB2 = 28
    TNS_DATA_TYPE_SB4 = 29
    TNS_DATA_TYPE_SWORD = 30
    TNS_DATA_TYPE_UWORD = 31
    TNS_DATA_TYPE_PTRB = 32
    TNS_DATA_TYPE_PTRW = 33
    TNS_DATA_TYPE_OER8 = 34 + 256
    TNS_DATA_TYPE_FUN = 35 + 256
    TNS_DATA_TYPE_AUA = 36 + 256
    TNS_DATA_TYPE_RXH7 = 37 + 256
    TNS_DATA_TYPE_NA6 = 38 + 256
    TNS_DATA_TYPE_OAC9 = 39
    TNS_DATA_TYPE_AMS = 40
    TNS_DATA_TYPE_BRN = 41
    TNS_DATA_TYPE_BRP = 42 + 256
    TNS_DATA_TYPE_BRV = 43 + 256
    TNS_DATA_TYPE_KVA = 44 + 256
    TNS_DATA_TYPE_CLS = 45 + 256
    TNS_DATA_TYPE_CUI = 46 + 256
    TNS_DATA_TYPE_DFN = 47 + 256
    TNS_DATA_TYPE_DQR = 48 + 256
    TNS_DATA_TYPE_DSC = 49 + 256
    TNS_DATA_TYPE_EXE = 50 + 256
    TNS_DATA_TYPE_FCH = 51 + 256
    TNS_DATA_TYPE_GBV = 52 + 256
    TNS_DATA_TYPE_GEM = 53 + 256
    TNS_DATA_TYPE_GIV = 54 + 256
    TNS_DATA_TYPE_OKG = 55 + 256
    TNS_DATA_TYPE_HMI = 56 + 256
    TNS_DATA_TYPE_INO = 57 + 256
    TNS_DATA_TYPE_LNF = 59 + 256
    TNS_DATA_TYPE_ONT = 60 + 256
    TNS_DATA_TYPE_OPE = 61 + 256
    TNS_DATA_TYPE_OSQ = 62 + 256
    TNS_DATA_TYPE_SFE = 63 + 256
    TNS_DATA_TYPE_SPF = 64 + 256
    TNS_DATA_TYPE_VSN = 65 + 256
    TNS_DATA_TYPE_UD7 = 66 + 256
    TNS_DATA_TYPE_DSA = 67 + 256
    TNS_DATA_TYPE_UIN = 68
    TNS_DATA_TYPE_PIN = 71 + 256
    TNS_DATA_TYPE_PFN = 72 + 256
    TNS_DATA_TYPE_PPT = 73 + 256
    TNS_DATA_TYPE_STO = 75 + 256
    TNS_DATA_TYPE_ARC = 77 + 256
    TNS_DATA_TYPE_MRS = 78 + 256
    TNS_DATA_TYPE_MRT = 79 + 256
    TNS_DATA_TYPE_MRG = 80 + 256
    TNS_DATA_TYPE_MRR = 81 + 256
    TNS_DATA_TYPE_MRC = 82 + 256
    TNS_DATA_TYPE_VER = 83 + 256
    TNS_DATA_TYPE_LON2 = 84 + 256
    TNS_DATA_TYPE_INO2 = 85 + 256
    TNS_DATA_TYPE_ALL = 86 + 256
    TNS_DATA_TYPE_UDB = 87 + 256
    TNS_DATA_TYPE_AQI = 88 + 256
    TNS_DATA_TYPE_ULB = 89 + 256
    TNS_DATA_TYPE_ULD = 90 + 256
    TNS_DATA_TYPE_SLS = 91
    TNS_DATA_TYPE_SID = 92 + 256
    TNS_DATA_TYPE_NA7 = 93 + 256
    TNS_DATA_TYPE_LVC = 94
    TNS_DATA_TYPE_LVB = 95
    TNS_DATA_TYPE_AVC = 97
    TNS_DATA_TYPE_AL7 = 98 + 256
    TNS_DATA_TYPE_K2RPC = 99 + 256
    TNS_DATA_TYPE_RDD = 104
    TNS_DATA_TYPE_XDP = 103 + 256
    TNS_DATA_TYPE_OSL = 106
    TNS_DATA_TYPE_OKO8 = 107 + 256
    TNS_DATA_TYPE_EXT_NAMED = 108
    TNS_DATA_TYPE_EXT_REF = 110
    TNS_DATA_TYPE_INT_REF = 111
    TNS_DATA_TYPE_CFILE = 115
    TNS_DATA_TYPE_RSET = 116
    TNS_DATA_TYPE_CWD = 117
    TNS_DATA_TYPE_OAC122 = 120
    TNS_DATA_TYPE_UD12 = 124 + 256
    TNS_DATA_TYPE_AL8 = 125 + 256
    TNS_DATA_TYPE_LFOP = 126 + 256
    TNS_DATA_TYPE_FCRT = 127 + 256
    TNS_DATA_TYPE_DNY = 128 + 256
    TNS_DATA_TYPE_OPR = 129 + 256
    TNS_DATA_TYPE_PLS = 130 + 256
    TNS_DATA_TYPE_XID = 131 + 256
    TNS_DATA_TYPE_TXN = 132 + 256
    TNS_DATA_TYPE_DCB = 133 + 256
    TNS_DATA_TYPE_CCA = 134 + 256
    TNS_DATA_TYPE_WRN = 135 + 256
    TNS_DATA_TYPE_TLH = 137 + 256
    TNS_DATA_TYPE_TOH = 138 + 256
    TNS_DATA_TYPE_FOI = 139 + 256
    TNS_DATA_TYPE_SID2 = 140 + 256
    TNS_DATA_TYPE_TCH = 141 + 256
    TNS_DATA_TYPE_PII = 142 + 256
    TNS_DATA_TYPE_PFI = 143 + 256
    TNS_DATA_TYPE_PPU = 144 + 256
    TNS_DATA_TYPE_PTE = 145 + 256
    TNS_DATA_TYPE_CLV = 146
    TNS_DATA_TYPE_RXH8 = 148 + 256
    TNS_DATA_TYPE_N12 = 149 + 256
    TNS_DATA_TYPE_AUTH = 150 + 256
    TNS_DATA_TYPE_KVAL = 151 + 256
    TNS_DATA_TYPE_DTR = 152
    TNS_DATA_TYPE_DUN = 153
    TNS_DATA_TYPE_DOP = 154
    TNS_DATA_TYPE_VST = 155
    TNS_DATA_TYPE_ODT = 156
    TNS_DATA_TYPE_FGI = 157 + 256
    TNS_DATA_TYPE_DSY = 158 + 256
    TNS_DATA_TYPE_DSYR8 = 159 + 256
    TNS_DATA_TYPE_DSYH8 = 160 + 256
    TNS_DATA_TYPE_DSYL = 161 + 256
    TNS_DATA_TYPE_DSYT8 = 162 + 256
    TNS_DATA_TYPE_DSYV8 = 163 + 256
    TNS_DATA_TYPE_DSYP = 164 + 256
    TNS_DATA_TYPE_DSYF = 165 + 256
    TNS_DATA_TYPE_DSYK = 166 + 256
    TNS_DATA_TYPE_DSYY = 167 + 256
    TNS_DATA_TYPE_DSYQ = 168 + 256
    TNS_DATA_TYPE_DSYC = 169 + 256
    TNS_DATA_TYPE_DSYA = 170 + 256
    TNS_DATA_TYPE_OT8 = 171 + 256
    TNS_DATA_TYPE_DOL = 172
    TNS_DATA_TYPE_DSYTY = 173 + 256
    TNS_DATA_TYPE_AQE = 174 + 256
    TNS_DATA_TYPE_KV = 175 + 256
    TNS_DATA_TYPE_AQD = 176 + 256
    TNS_DATA_TYPE_AQ8 = 177 + 256
    TNS_DATA_TYPE_TIME = 178
    TNS_DATA_TYPE_TIME_TZ = 179
    TNS_DATA_TYPE_EDATE = 184
    TNS_DATA_TYPE_ETIME = 185
    TNS_DATA_TYPE_ETTZ = 186
    TNS_DATA_TYPE_ESTAMP = 187
    TNS_DATA_TYPE_ESTZ = 188
    TNS_DATA_TYPE_EIYM = 189
    TNS_DATA_TYPE_EIDS = 190
    TNS_DATA_TYPE_RFS = 193 + 256
    TNS_DATA_TYPE_RXH10 = 194 + 256
    TNS_DATA_TYPE_DCLOB = 195
    TNS_DATA_TYPE_DBLOB = 196
    TNS_DATA_TYPE_DBFILE = 197
    TNS_DATA_TYPE_DJSON = 198
    TNS_DATA_TYPE_KPN = 198 + 256
    TNS_DATA_TYPE_KPDNR = 199 + 256
    TNS_DATA_TYPE_DSYD = 200 + 256
    TNS_DATA_TYPE_DSYS = 201 + 256
    TNS_DATA_TYPE_DSYR = 202 + 256
    TNS_DATA_TYPE_DSYH = 203 + 256
    TNS_DATA_TYPE_DSYT = 204 + 256
    TNS_DATA_TYPE_DSYV = 205 + 256
    TNS_DATA_TYPE_AQM = 206 + 256
    TNS_DATA_TYPE_OER11 = 207 + 256
    TNS_DATA_TYPE_AQL = 210 + 256
    TNS_DATA_TYPE_OTC = 211 + 256
    TNS_DATA_TYPE_KFNO = 212 + 256
    TNS_DATA_TYPE_KFNP = 213 + 256
    TNS_DATA_TYPE_KGT8 = 214 + 256
    TNS_DATA_TYPE_RASB4 = 215 + 256
    TNS_DATA_TYPE_RAUB2 = 216 + 256
    TNS_DATA_TYPE_RAUB1 = 217 + 256
    TNS_DATA_TYPE_RATXT = 218 + 256
    TNS_DATA_TYPE_RSSB4 = 219 + 256
    TNS_DATA_TYPE_RSUB2 = 220 + 256
    TNS_DATA_TYPE_RSUB1 = 221 + 256
    TNS_DATA_TYPE_RSTXT = 222 + 256
    TNS_DATA_TYPE_RIDL = 223 + 256
    TNS_DATA_TYPE_GLRDD = 224 + 256
    TNS_DATA_TYPE_GLRDG = 225 + 256
    TNS_DATA_TYPE_GLRDC = 226 + 256
    TNS_DATA_TYPE_OKO = 227 + 256
    TNS_DATA_TYPE_DPP = 228 + 256
    TNS_DATA_TYPE_DPLS = 229 + 256
    TNS_DATA_TYPE_DPMOP = 230 + 256
    TNS_DATA_TYPE_ESITZ = 232
    TNS_DATA_TYPE_UB8 = 233
    TNS_DATA_TYPE_STAT = 234 + 256
    TNS_DATA_TYPE_RFX = 235 + 256
    TNS_DATA_TYPE_FAL = 236 + 256
    TNS_DATA_TYPE_CKV = 237 + 256
    TNS_DATA_TYPE_DRCX = 238 + 256
    TNS_DATA_TYPE_KGH = 239 + 256
    TNS_DATA_TYPE_AQO = 240 + 256
    TNS_DATA_TYPE_PNTY = 241
    TNS_DATA_TYPE_OKGT = 242 + 256
    TNS_DATA_TYPE_KPFC = 243 + 256
    TNS_DATA_TYPE_FE2 = 244 + 256
    TNS_DATA_TYPE_SPFP = 245 + 256
    TNS_DATA_TYPE_DPULS = 246 + 256
    TNS_DATA_TYPE_AQA = 253 + 256
    TNS_DATA_TYPE_KPBF = 254 + 256
    TNS_DATA_TYPE_TSM = 513
    TNS_DATA_TYPE_MSS = 514
    TNS_DATA_TYPE_KPC = 516
    TNS_DATA_TYPE_CRS = 517
    TNS_DATA_TYPE_KKS = 518
    TNS_DATA_TYPE_KSP = 519
    TNS_DATA_TYPE_KSPTOP = 520
    TNS_DATA_TYPE_KSPVAL = 521
    TNS_DATA_TYPE_PSS = 522
    TNS_DATA_TYPE_NLS = 523
    TNS_DATA_TYPE_ALS = 524
    TNS_DATA_TYPE_KSDEVTVAL = 525
    TNS_DATA_TYPE_KSDEVTTOP = 526
    TNS_DATA_TYPE_KPSPP = 527
    TNS_DATA_TYPE_KOL = 528
    TNS_DATA_TYPE_LST = 529
    TNS_DATA_TYPE_ACX = 530
    TNS_DATA_TYPE_SCS = 531
    TNS_DATA_TYPE_RXH = 532
    TNS_DATA_TYPE_KPDNS = 533
    TNS_DATA_TYPE_KPDCN = 534
    TNS_DATA_TYPE_KPNNS = 535
    TNS_DATA_TYPE_KPNCN = 536
    TNS_DATA_TYPE_KPS = 537
    TNS_DATA_TYPE_APINF = 538
    TNS_DATA_TYPE_TEN = 539
    TNS_DATA_TYPE_XSSCS = 540
    TNS_DATA_TYPE_XSSSO = 541
    TNS_DATA_TYPE_XSSAO = 542
    TNS_DATA_TYPE_KSRPC = 543
    TNS_DATA_TYPE_KVL = 560
    TNS_DATA_TYPE_SESSGET = 563
    TNS_DATA_TYPE_SESSREL = 564
    TNS_DATA_TYPE_XSSDEF = 565
    TNS_DATA_TYPE_PDQCINV = 572
    TNS_DATA_TYPE_PDQIDC = 573
    TNS_DATA_TYPE_KPDQCSTA = 574
    TNS_DATA_TYPE_KPRS = 575
    TNS_DATA_TYPE_KPDQIDC = 576
    TNS_DATA_TYPE_RTSTRM = 578
    TNS_DATA_TYPE_SESSRET = 579
    TNS_DATA_TYPE_SCN6 = 580
    TNS_DATA_TYPE_KECPA = 581
    TNS_DATA_TYPE_KECPP = 582
    TNS_DATA_TYPE_SXA = 583
    TNS_DATA_TYPE_KVARR = 584
    TNS_DATA_TYPE_KPNGN = 585
    TNS_DATA_TYPE_XSNSOP = 590
    TNS_DATA_TYPE_XSATTR = 591
    TNS_DATA_TYPE_XSNS = 592
    TNS_DATA_TYPE_TXT = 593
    TNS_DATA_TYPE_XSSESSNS = 594
    TNS_DATA_TYPE_XSATTOP = 595
    TNS_DATA_TYPE_XSCREOP = 596
    TNS_DATA_TYPE_XSDETOP = 597
    TNS_DATA_TYPE_XSDESOP = 598
    TNS_DATA_TYPE_XSSETSP = 599
    TNS_DATA_TYPE_XSSIDP = 600
    TNS_DATA_TYPE_XSPRIN = 601
    TNS_DATA_TYPE_XSKVL = 602
    TNS_DATA_TYPE_XSSSDEF2 = 603
    TNS_DATA_TYPE_XSNSOP2 = 604
    TNS_DATA_TYPE_XSNS2 = 605
    TNS_DATA_TYPE_IMPLRES = 611
    TNS_DATA_TYPE_OER19 = 612
    TNS_DATA_TYPE_UB1ARRAY = 613
    TNS_DATA_TYPE_SESSSTATE = 614
    TNS_DATA_TYPE_AC_REPLAY = 615
    TNS_DATA_TYPE_AC_CONT = 616
    TNS_DATA_TYPE_KPDNREQ = 622
    TNS_DATA_TYPE_KPDNRNF = 623
    TNS_DATA_TYPE_KPNGNC = 624
    TNS_DATA_TYPE_KPNRI = 625
    TNS_DATA_TYPE_AQENQ = 626
    TNS_DATA_TYPE_AQDEQ = 627
    TNS_DATA_TYPE_AQJMS = 628
    TNS_DATA_TYPE_KPDNRPAY = 629
    TNS_DATA_TYPE_KPDNRACK = 630
    TNS_DATA_TYPE_KPDNRMP = 631
    TNS_DATA_TYPE_KPDNRDQ = 632
    TNS_DATA_TYPE_CHUNKINFO = 636
    TNS_DATA_TYPE_SCN = 637
    TNS_DATA_TYPE_SCN8 = 638
    TNS_DATA_TYPE_UD21 = 639
    TNS_DATA_TYPE_TNP = 640
    TNS_DATA_TYPE_OAC = 646
    TNS_DATA_TYPE_SESSSIGN = 647
    TNS_DATA_TYPE_OER = 652
    TNS_DATA_TYPE_PLEND = 660
    TNS_DATA_TYPE_PLBGN = 661
    TNS_DATA_TYPE_UDS = 663
    TNS_DATA_TYPE_PLOP = 665


# data type representations
cdef enum:
    TNS_TYPE_REP_NATIVE = 0
    TNS_TYPE_REP_UNIVERSAL = 1
    TNS_TYPE_REP_ORACLE = 10


cdef DataType[320] DATA_TYPES = [
    [ORA_TYPE_NUM_VARCHAR, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_NUMBER, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [ORA_TYPE_NUM_LONG, ORA_TYPE_NUM_LONG, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_DATE, ORA_TYPE_NUM_DATE, TNS_TYPE_REP_ORACLE],
    [ORA_TYPE_NUM_RAW, ORA_TYPE_NUM_RAW, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_LONG_RAW, ORA_TYPE_NUM_LONG_RAW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB2, TNS_DATA_TYPE_UB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB4, TNS_DATA_TYPE_UB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SB1, TNS_DATA_TYPE_SB1, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_SB2, TNS_DATA_TYPE_SB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SB4, TNS_DATA_TYPE_SB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SWORD, TNS_DATA_TYPE_SWORD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UWORD, TNS_DATA_TYPE_UWORD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTRB, TNS_DATA_TYPE_PTRB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTRW, TNS_DATA_TYPE_PTRW, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIDDEF, TNS_DATA_TYPE_TIDDEF, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_ROWID, ORA_TYPE_NUM_ROWID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AMS, TNS_DATA_TYPE_AMS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRN, TNS_DATA_TYPE_BRN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CWD, TNS_DATA_TYPE_CWD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OAC122, TNS_DATA_TYPE_OAC122, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER8, TNS_DATA_TYPE_OER8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FUN, TNS_DATA_TYPE_FUN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AUA, TNS_DATA_TYPE_AUA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH7, TNS_DATA_TYPE_RXH7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NA6, TNS_DATA_TYPE_NA6, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRP, TNS_DATA_TYPE_BRP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_BRV, TNS_DATA_TYPE_BRV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVA, TNS_DATA_TYPE_KVA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CLS, TNS_DATA_TYPE_CLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CUI, TNS_DATA_TYPE_CUI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DFN, TNS_DATA_TYPE_DFN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DQR, TNS_DATA_TYPE_DQR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSC, TNS_DATA_TYPE_DSC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXE, TNS_DATA_TYPE_EXE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FCH, TNS_DATA_TYPE_FCH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GBV, TNS_DATA_TYPE_GBV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GEM, TNS_DATA_TYPE_GEM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GIV, TNS_DATA_TYPE_GIV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKG, TNS_DATA_TYPE_OKG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_HMI, TNS_DATA_TYPE_HMI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INO, TNS_DATA_TYPE_INO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LNF, TNS_DATA_TYPE_LNF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ONT, TNS_DATA_TYPE_ONT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OPE, TNS_DATA_TYPE_OPE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OSQ, TNS_DATA_TYPE_OSQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SFE, TNS_DATA_TYPE_SFE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SPF, TNS_DATA_TYPE_SPF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VSN, TNS_DATA_TYPE_VSN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UD7, TNS_DATA_TYPE_UD7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSA, TNS_DATA_TYPE_DSA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PIN, TNS_DATA_TYPE_PIN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PFN, TNS_DATA_TYPE_PFN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PPT, TNS_DATA_TYPE_PPT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_STO, TNS_DATA_TYPE_STO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ARC, TNS_DATA_TYPE_ARC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRS, TNS_DATA_TYPE_MRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRT, TNS_DATA_TYPE_MRT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRG, TNS_DATA_TYPE_MRG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRR, TNS_DATA_TYPE_MRR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MRC, TNS_DATA_TYPE_MRC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VER, TNS_DATA_TYPE_VER, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LON2, TNS_DATA_TYPE_LON2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INO2, TNS_DATA_TYPE_INO2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ALL, TNS_DATA_TYPE_ALL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UDB, TNS_DATA_TYPE_UDB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQI, TNS_DATA_TYPE_AQI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ULB, TNS_DATA_TYPE_ULB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ULD, TNS_DATA_TYPE_ULD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SID, TNS_DATA_TYPE_SID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NA7, TNS_DATA_TYPE_NA7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AL7, TNS_DATA_TYPE_AL7, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_K2RPC, TNS_DATA_TYPE_K2RPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XDP, TNS_DATA_TYPE_XDP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKO8, TNS_DATA_TYPE_OKO8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UD12, TNS_DATA_TYPE_UD12, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AL8, TNS_DATA_TYPE_AL8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LFOP, TNS_DATA_TYPE_LFOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FCRT, TNS_DATA_TYPE_FCRT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DNY, TNS_DATA_TYPE_DNY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OPR, TNS_DATA_TYPE_OPR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PLS, TNS_DATA_TYPE_PLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XID, TNS_DATA_TYPE_XID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TXN, TNS_DATA_TYPE_TXN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DCB, TNS_DATA_TYPE_DCB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CCA, TNS_DATA_TYPE_CCA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_WRN, TNS_DATA_TYPE_WRN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TLH, TNS_DATA_TYPE_TLH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TOH, TNS_DATA_TYPE_TOH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FOI, TNS_DATA_TYPE_FOI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SID2, TNS_DATA_TYPE_SID2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TCH, TNS_DATA_TYPE_TCH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PII, TNS_DATA_TYPE_PII, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PFI, TNS_DATA_TYPE_PFI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PPU, TNS_DATA_TYPE_PPU, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PTE, TNS_DATA_TYPE_PTE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH8, TNS_DATA_TYPE_RXH8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_N12, TNS_DATA_TYPE_N12, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AUTH, TNS_DATA_TYPE_AUTH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVAL, TNS_DATA_TYPE_KVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FGI, TNS_DATA_TYPE_FGI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSY, TNS_DATA_TYPE_DSY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYR8, TNS_DATA_TYPE_DSYR8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYH8, TNS_DATA_TYPE_DSYH8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYL, TNS_DATA_TYPE_DSYL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYT8, TNS_DATA_TYPE_DSYT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYV8, TNS_DATA_TYPE_DSYV8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYP, TNS_DATA_TYPE_DSYP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYF, TNS_DATA_TYPE_DSYF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYK, TNS_DATA_TYPE_DSYK, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYY, TNS_DATA_TYPE_DSYY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYQ, TNS_DATA_TYPE_DSYQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYC, TNS_DATA_TYPE_DSYC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYA, TNS_DATA_TYPE_DSYA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OT8, TNS_DATA_TYPE_OT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYTY, TNS_DATA_TYPE_DSYTY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQE, TNS_DATA_TYPE_AQE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KV, TNS_DATA_TYPE_KV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQD, TNS_DATA_TYPE_AQD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQ8, TNS_DATA_TYPE_AQ8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RFS, TNS_DATA_TYPE_RFS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH10, TNS_DATA_TYPE_RXH10, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPN, TNS_DATA_TYPE_KPN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNR, TNS_DATA_TYPE_KPDNR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYD, TNS_DATA_TYPE_DSYD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYS, TNS_DATA_TYPE_DSYS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYR, TNS_DATA_TYPE_DSYR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYH, TNS_DATA_TYPE_DSYH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYT, TNS_DATA_TYPE_DSYT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DSYV, TNS_DATA_TYPE_DSYV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQM, TNS_DATA_TYPE_AQM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER11, TNS_DATA_TYPE_OER11, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQL, TNS_DATA_TYPE_AQL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OTC, TNS_DATA_TYPE_OTC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KFNO, TNS_DATA_TYPE_KFNO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KFNP, TNS_DATA_TYPE_KFNP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KGT8, TNS_DATA_TYPE_KGT8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RASB4, TNS_DATA_TYPE_RASB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RAUB2, TNS_DATA_TYPE_RAUB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RAUB1, TNS_DATA_TYPE_RAUB1, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RATXT, TNS_DATA_TYPE_RATXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSSB4, TNS_DATA_TYPE_RSSB4, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSUB2, TNS_DATA_TYPE_RSUB2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSUB1, TNS_DATA_TYPE_RSUB1, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSTXT, TNS_DATA_TYPE_RSTXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RIDL, TNS_DATA_TYPE_RIDL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDD, TNS_DATA_TYPE_GLRDD, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDG, TNS_DATA_TYPE_GLRDG, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_GLRDC, TNS_DATA_TYPE_GLRDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKO, TNS_DATA_TYPE_OKO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPP, TNS_DATA_TYPE_DPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPLS, TNS_DATA_TYPE_DPLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPMOP, TNS_DATA_TYPE_DPMOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_STAT, TNS_DATA_TYPE_STAT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RFX, TNS_DATA_TYPE_RFX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FAL, TNS_DATA_TYPE_FAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CKV, TNS_DATA_TYPE_CKV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DRCX, TNS_DATA_TYPE_DRCX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KGH, TNS_DATA_TYPE_KGH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQO, TNS_DATA_TYPE_AQO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OKGT, TNS_DATA_TYPE_OKGT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPFC, TNS_DATA_TYPE_KPFC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_FE2, TNS_DATA_TYPE_FE2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SPFP, TNS_DATA_TYPE_SPFP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DPULS, TNS_DATA_TYPE_DPULS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQA, TNS_DATA_TYPE_AQA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPBF, TNS_DATA_TYPE_KPBF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TSM, TNS_DATA_TYPE_TSM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_MSS, TNS_DATA_TYPE_MSS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPC, TNS_DATA_TYPE_KPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CRS, TNS_DATA_TYPE_CRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KKS, TNS_DATA_TYPE_KKS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSP, TNS_DATA_TYPE_KSP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSPTOP, TNS_DATA_TYPE_KSPTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSPVAL, TNS_DATA_TYPE_KSPVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PSS, TNS_DATA_TYPE_PSS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_NLS, TNS_DATA_TYPE_NLS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ALS, TNS_DATA_TYPE_ALS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSDEVTVAL, TNS_DATA_TYPE_KSDEVTVAL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSDEVTTOP, TNS_DATA_TYPE_KSDEVTTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPSPP, TNS_DATA_TYPE_KPSPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KOL, TNS_DATA_TYPE_KOL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LST, TNS_DATA_TYPE_LST, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ACX, TNS_DATA_TYPE_ACX, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCS, TNS_DATA_TYPE_SCS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RXH, TNS_DATA_TYPE_RXH, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNS, TNS_DATA_TYPE_KPDNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDCN, TNS_DATA_TYPE_KPDCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNNS, TNS_DATA_TYPE_KPNNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNCN, TNS_DATA_TYPE_KPNCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPS, TNS_DATA_TYPE_KPS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_APINF, TNS_DATA_TYPE_APINF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TEN, TNS_DATA_TYPE_TEN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSCS, TNS_DATA_TYPE_XSSCS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSSO, TNS_DATA_TYPE_XSSSO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSAO, TNS_DATA_TYPE_XSSAO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KSRPC, TNS_DATA_TYPE_KSRPC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVL, TNS_DATA_TYPE_KVL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSDEF, TNS_DATA_TYPE_XSSDEF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PDQCINV, TNS_DATA_TYPE_PDQCINV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PDQIDC, TNS_DATA_TYPE_PDQIDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDQCSTA, TNS_DATA_TYPE_KPDQCSTA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPRS, TNS_DATA_TYPE_KPRS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDQIDC, TNS_DATA_TYPE_KPDQIDC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RTSTRM, TNS_DATA_TYPE_RTSTRM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSGET, TNS_DATA_TYPE_SESSGET, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSREL, TNS_DATA_TYPE_SESSREL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSRET, TNS_DATA_TYPE_SESSRET, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN6, TNS_DATA_TYPE_SCN6, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KECPA, TNS_DATA_TYPE_KECPA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KECPP, TNS_DATA_TYPE_KECPP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SXA, TNS_DATA_TYPE_SXA, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KVARR, TNS_DATA_TYPE_KVARR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNGN, TNS_DATA_TYPE_KPNGN, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BINARY_INTEGER, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_FLOAT, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_STR, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VNU, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_PDN, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_VCS, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_VBI, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OAC9, TNS_DATA_TYPE_OAC9, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UIN, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_SLS, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_LVC, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_LVB, ORA_TYPE_NUM_RAW, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_CHAR, ORA_TYPE_NUM_CHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AVC, ORA_TYPE_NUM_CHAR, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BINARY_FLOAT, ORA_TYPE_NUM_BINARY_FLOAT,
            TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BINARY_DOUBLE, ORA_TYPE_NUM_BINARY_DOUBLE,
            TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_CURSOR, ORA_TYPE_NUM_CURSOR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RDD, ORA_TYPE_NUM_ROWID, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OSL, TNS_DATA_TYPE_OSL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXT_NAMED, ORA_TYPE_NUM_OBJECT, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_OBJECT, ORA_TYPE_NUM_OBJECT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EXT_REF, TNS_DATA_TYPE_INT_REF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_INT_REF, TNS_DATA_TYPE_INT_REF, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_CLOB, ORA_TYPE_NUM_CLOB, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BLOB, ORA_TYPE_NUM_BLOB, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BFILE, ORA_TYPE_NUM_BFILE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CFILE, TNS_DATA_TYPE_CFILE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_RSET, ORA_TYPE_NUM_CURSOR, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_JSON, ORA_TYPE_NUM_JSON, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DJSON, TNS_DATA_TYPE_DJSON, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CLV, TNS_DATA_TYPE_CLV, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DTR, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DUN, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DOP, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_VST, ORA_TYPE_NUM_VARCHAR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ODT, ORA_TYPE_NUM_DATE, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_DOL, ORA_TYPE_NUM_NUMBER, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_TIME, TNS_DATA_TYPE_TIME, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TIME_TZ, TNS_DATA_TYPE_TIME_TZ, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_TIMESTAMP, ORA_TYPE_NUM_TIMESTAMP, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_TIMESTAMP_TZ, ORA_TYPE_NUM_TIMESTAMP_TZ,
            TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_INTERVAL_YM, ORA_TYPE_NUM_INTERVAL_YM,
            TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_INTERVAL_DS, ORA_TYPE_NUM_INTERVAL_DS,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EDATE, ORA_TYPE_NUM_DATE, TNS_TYPE_REP_ORACLE],
    [TNS_DATA_TYPE_ETIME, TNS_DATA_TYPE_ETIME, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ETTZ, TNS_DATA_TYPE_ETTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESTAMP, TNS_DATA_TYPE_ESTAMP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESTZ, TNS_DATA_TYPE_ESTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EIYM, TNS_DATA_TYPE_EIYM, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_EIDS, TNS_DATA_TYPE_EIDS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DCLOB, ORA_TYPE_NUM_CLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DBLOB, ORA_TYPE_NUM_BLOB, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_DBFILE, ORA_TYPE_NUM_BFILE, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_UROWID, ORA_TYPE_NUM_UROWID, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_TIMESTAMP_LTZ, ORA_TYPE_NUM_TIMESTAMP_LTZ,
            TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_ESITZ, ORA_TYPE_NUM_TIMESTAMP_LTZ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB8, TNS_DATA_TYPE_UB8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PNTY, ORA_TYPE_NUM_OBJECT, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_BOOLEAN, ORA_TYPE_NUM_BOOLEAN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNSOP, TNS_DATA_TYPE_XSNSOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSATTR, TNS_DATA_TYPE_XSATTR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNS, TNS_DATA_TYPE_XSNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UB1ARRAY, TNS_DATA_TYPE_UB1ARRAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSSTATE, TNS_DATA_TYPE_SESSSTATE, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AC_REPLAY, TNS_DATA_TYPE_AC_REPLAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AC_CONT, TNS_DATA_TYPE_AC_CONT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_IMPLRES, TNS_DATA_TYPE_IMPLRES, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER19, TNS_DATA_TYPE_OER19, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TXT, TNS_DATA_TYPE_TXT, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSESSNS, TNS_DATA_TYPE_XSSESSNS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSATTOP, TNS_DATA_TYPE_XSATTOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSCREOP, TNS_DATA_TYPE_XSCREOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSDETOP, TNS_DATA_TYPE_XSDETOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSDESOP, TNS_DATA_TYPE_XSDESOP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSETSP, TNS_DATA_TYPE_XSSETSP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSIDP, TNS_DATA_TYPE_XSSIDP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSPRIN, TNS_DATA_TYPE_XSPRIN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSKVL, TNS_DATA_TYPE_XSKVL, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSSSDEF2, TNS_DATA_TYPE_XSSSDEF2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNSOP2, TNS_DATA_TYPE_XSNSOP2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_XSNS2, TNS_DATA_TYPE_XSNS2, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNREQ, TNS_DATA_TYPE_KPDNREQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRNF, TNS_DATA_TYPE_KPDNRNF, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNGNC, TNS_DATA_TYPE_KPNGNC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPNRI, TNS_DATA_TYPE_KPNRI, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQENQ, TNS_DATA_TYPE_AQENQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQDEQ, TNS_DATA_TYPE_AQDEQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_AQJMS, TNS_DATA_TYPE_AQJMS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRPAY, TNS_DATA_TYPE_KPDNRPAY, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRACK, TNS_DATA_TYPE_KPDNRACK, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRMP, TNS_DATA_TYPE_KPDNRMP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_KPDNRDQ, TNS_DATA_TYPE_KPDNRDQ, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN, TNS_DATA_TYPE_SCN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SCN8, TNS_DATA_TYPE_SCN8, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_CHUNKINFO, TNS_DATA_TYPE_CHUNKINFO, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UD21, TNS_DATA_TYPE_UD21, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_UDS, TNS_DATA_TYPE_UDS, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_TNP, TNS_DATA_TYPE_TNP, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OER, TNS_DATA_TYPE_OER, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_OAC, TNS_DATA_TYPE_OAC, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_SESSSIGN, TNS_DATA_TYPE_SESSSIGN, TNS_TYPE_REP_UNIVERSAL],
    [ORA_TYPE_NUM_VECTOR, ORA_TYPE_NUM_VECTOR, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PLEND, TNS_DATA_TYPE_PLEND, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PLBGN, TNS_DATA_TYPE_PLBGN, TNS_TYPE_REP_UNIVERSAL],
    [TNS_DATA_TYPE_PLOP, TNS_DATA_TYPE_PLOP, TNS_TYPE_REP_UNIVERSAL],
    [0, 0, 0]
]


@cython.final
cdef class DataTypesMessage(Message):

    cdef int _process_message(self, ReadBuffer buf,
                              uint8_t message_type) except -1:
        cdef uint16_t data_type, conv_data_type
        while True:
            buf.read_uint16be(&data_type)
            if data_type == 0:
                break
            buf.read_uint16be(&conv_data_type)
            if conv_data_type != 0:
                buf.skip_raw_bytes(4)
        if not buf._caps.supports_end_of_response:
            self.end_of_response = True

    cdef int _write_message(self, WriteBuffer buf) except -1:
        cdef:
            DataType* data_type
            int i

        # write character set and capabilities
        buf.write_uint8(TNS_MSG_TYPE_DATA_TYPES)
        buf.write_uint16le(TNS_CHARSET_UTF8)
        buf.write_uint16le(TNS_CHARSET_UTF8)
        buf.write_uint8(TNS_ENCODING_MULTI_BYTE | TNS_ENCODING_CONV_LENGTH)
        buf.write_bytes_with_length(bytes(buf._caps.compile_caps))
        buf.write_bytes_with_length(bytes(buf._caps.runtime_caps))

        # write data types
        i = 0
        while True:
            data_type = &DATA_TYPES[i]
            if data_type.data_type == 0:
                break
            i += 1
            buf.write_uint16be(data_type.data_type)
            buf.write_uint16be(data_type.conv_data_type)
            buf.write_uint16be(data_type.representation)
            buf.write_uint16be(0)
        buf.write_uint16be(0)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# azure_tokens.py
#
# Methods that generates an OAuth2 access token using the MSAL SDK
# -----------------------------------------------------------------------------

import msal
import oracledb


def generate_token(token_auth_config, refresh=False):
    """
    Generates an Azure access token based on provided credentials.
    """
    user_auth_type = token_auth_config.get("auth_type") or ""
    auth_type = user_auth_type.lower()
    if auth_type == "azureserviceprincipal":
        return _service_principal_credentials(token_auth_config)
    else:
        raise ValueError(
            f"Unrecognized auth_type authentication method: {user_auth_type}"
        )


def _service_principal_credentials(token_auth_config):
    """
    Returns the access token for authentication as a service principal.
    """
    msal_config = {
        "authority": token_auth_config["authority"],
        "client_id": token_auth_config["client_id"],
        "client_credential": token_auth_config["client_credential"],
    }
    # Initialize the Confidential Client Application
    cca = msal.ConfidentialClientApplication(**msal_config)
    auth_response = cca.acquire_token_for_client(
        scopes=[token_auth_config["scopes"]]
    )

    if "access_token" in auth_response:
        return auth_response["access_token"]


def azure_token_hook(params: oracledb.ConnectParams):
    """
    Azure-specific hook for generating a token.
    """
    if params.extra_auth_params is not None:

        def token_callback(refresh):
            return generate_token(params.extra_auth_params, refresh)

        params.set(access_token=token_callback)


# Register the token hook for Azure
oracledb.register_params_hook(azure_token_hook)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3900 - Module for testing the cursor execute() method
"""

import collections

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def test_3900(self):
        "3900 - test executing a statement without any arguments"
        result = self.cursor.execute("begin null; end;")
        self.assertIsNone(result)

    def test_3901(self):
        "3901 - test executing a None statement with bind variables"
        cursor = self.conn.cursor()
        with self.assertRaisesFullCode("DPY-2001"):
            cursor.execute(None, x=5)

    def test_3902(self):
        "3902 - test executing a statement with args and empty keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        args = [simple_var]
        kwargs = {}
        result = self.cursor.execute("begin :1 := 25; end;", args, **kwargs)
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 25)

    def test_3903(self):
        "3903 - test executing a statement with keyword arguments"
        simple_var = self.cursor.var(oracledb.NUMBER)
        result = self.cursor.execute(
            "begin :value := 5; end;", value=simple_var
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 5)

    def test_3904(self):
        "3904 - test executing a statement with a dictionary argument"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        result = self.cursor.execute("begin :value := 10; end;", dict_arg)
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3905(self):
        "3905 - test executing a statement with both a dict and keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        with self.assertRaisesFullCode("DPY-2005"):
            self.cursor.execute(
                "begin :value := 15; end;", dict_arg, value=simple_var
            )

    def test_3906(self):
        "3906 - test executing a statement and then changing the array size"
        self.cursor.execute("select IntCol from TestNumbers")
        self.cursor.arraysize = 20
        self.assertEqual(len(self.cursor.fetchall()), 10)

    def test_3907(self):
        "3907 - test that subsequent executes succeed after bad execute"
        sql = "begin raise_application_error(-20000, 'this); end;"
        with self.assertRaisesFullCode("DPY-2041"):
            self.cursor.execute(sql)
        self.cursor.execute("begin null; end;")

    def test_3908(self):
        "3908 - test that subsequent fetches fail after bad execute"
        with self.assertRaisesFullCode("ORA-00904"):
            self.cursor.execute("select y from dual")
        with self.assertRaisesFullCode("DPY-1003"):
            self.cursor.fetchall()

    def test_3909(self):
        "3909 - test executing a statement with an incorrect named bind"
        sql = "select * from TestStrings where IntCol = :value"
        with self.assertRaisesFullCode("DPY-4008", "ORA-01036"):
            self.cursor.execute(sql, value2=3)

    def test_3910(self):
        "3910 - test executing a statement with named binds"
        result = self.cursor.execute(
            """
            select *
            from TestNumbers
            where IntCol = :value1 and LongIntCol = :value2
            """,
            value1=1,
            value2=38,
        )
        self.assertEqual(len(result.fetchall()), 1)

    def test_3911(self):
        "3911 - test executing a statement with an incorrect positional bind"
        sql = """
                select *
                from TestNumbers
                where IntCol = :value and LongIntCol = :value2"""
        with self.assertRaisesFullCode("DPY-4009", "ORA-01008"):
            self.cursor.execute(sql, [3])

    def test_3912(self):
        "3912 - test executing a statement with positional binds"
        result = self.cursor.execute(
            """
            select *
            from TestNumbers
            where IntCol = :value and LongIntCol = :value2
            """,
            [1, 38],
        )
        self.assertEqual(len(result.fetchall()), 1)

    def test_3913(self):
        "3913 - test executing a statement after rebinding a named bind"
        statement = "begin :value := :value2 + 5; end;"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var2 = self.cursor.var(oracledb.NUMBER)
        simple_var2.setvalue(0, 5)
        result = self.cursor.execute(
            statement, value=simple_var, value2=simple_var2
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

        simple_var = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2 = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2.setvalue(0, 10)
        result = self.cursor.execute(
            statement, value=simple_var, value2=simple_var2
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 15)

    def test_3914(self):
        "3914 - test executing a PL/SQL statement with duplicate binds"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        result = self.cursor.execute(
            """
            begin
                :value := :value + 5;
            end;
            """,
            value=simple_var,
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3915(self):
        "3915 - test executing a PL/SQL statement with duplicate binds"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        self.cursor.execute("begin :value := :value + 5; end;", [simple_var])
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3916(self):
        "3916 - test executing a statement with an incorrect number of binds"
        statement = "begin :value := :value2 + 5; end;"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, 5)
        with self.assertRaisesFullCode("DPY-4010", "ORA-01008"):
            self.cursor.execute(statement)
        with self.assertRaisesFullCode("DPY-4010", "ORA-01008"):
            self.cursor.execute(statement, value=var)
        with self.assertRaisesFullCode("DPY-4008", "ORA-01036"):
            self.cursor.execute(statement, value=var, value2=var, value3=var)

    def test_3917(self):
        "3917 - change in size on subsequent binds does not use optimised path"
        self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test String #1"), (2, "ABC" * 100)]
        for row in data:
            self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                """,
                row,
            )
        self.conn.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(), data)

    def test_3918(self):
        "3918 - test that dml can use optimised path"
        data_to_insert = [(i + 1, f"Test String #{i + 1}") for i in range(3)]
        self.cursor.execute("truncate table TestTempTable")
        for row in data_to_insert:
            with self.conn.cursor() as cursor:
                cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)
                    """,
                    row,
                )
        self.conn.commit()
        self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        self.assertEqual(self.cursor.fetchall(), data_to_insert)

    def test_3919(self):
        "3919 - test calling execute() with invalid parameters"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        with self.assertRaisesFullCode("DPY-2003"):
            self.cursor.execute(sql, "These are not valid parameters")

    def test_3920(self):
        "3920 - test calling execute() with mixed binds"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.setinputsizes(None, None, str)
        data = dict(val1=1, val2="Test String 1")
        with self.assertRaisesFullCode("DPY-2006"):
            self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                returning StringCol1 into :out_var
                """,
                data,
            )

    def test_3921(self):
        "3921 - test binding by name with double quotes"
        data = {'"_value1"': 1, '"VaLue_2"': 2, '"3VALUE"': 3}
        self.cursor.execute(
            'select :"_value1" + :"VaLue_2" + :"3VALUE" from dual',
            data,
        )
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 6)

    def test_3922(self):
        "3922 - test executing a statement with different input buffer sizes"
        sql = """
                insert into TestTempTable (IntCol, StringCol1, StringCol2)
                values (:int_col, :str_val1, :str_val2) returning IntCol
                into :ret_data"""
        values1 = {"int_col": 1, "str_val1": '{"a", "b"}', "str_val2": None}
        values2 = {"int_col": 2, "str_val1": None, "str_val2": '{"a", "b"}'}
        values3 = {"int_col": 3, "str_val1": '{"a"}', "str_val2": None}

        self.cursor.execute("truncate table TestTempTable")
        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values1)
        self.assertEqual(ret_bind.values, [["1"]])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values2)
        self.assertEqual(ret_bind.values, [["2"]])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values3)
        self.assertEqual(ret_bind.values, [["3"]])

    def test_3923(self):
        "3923 - test using rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Test 1')
            """
        )
        self.conn.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]

        def rowfactory(*row):
            return dict(zip(column_names, row))

        self.cursor.rowfactory = rowfactory
        self.assertEqual(self.cursor.rowfactory, rowfactory)
        self.assertEqual(
            self.cursor.fetchall(), [{"INTCOL": 1, "STRINGCOL1": "Test 1"}]
        )

    def test_3924(self):
        "3924 - test executing same query after setting rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test 1"), (2, "Test 2")]
        self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        self.conn.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        results1 = self.cursor.fetchall()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        results2 = self.cursor.fetchall()
        self.assertEqual(results1, results2)

    def test_3925(self):
        "3925 - test executing different query after setting rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test 1"), (2, "Test 2")]
        self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        self.conn.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        self.cursor.execute(
            """
            select IntCol, StringCol
            from TestSTrings
            where IntCol between 1 and 3 order by IntCol
            """
        )
        expected_data = [(1, "String 1"), (2, "String 2"), (3, "String 3")]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_3926(self):
        "3926 - test setting rowfactory on a REF cursor"
        with self.conn.cursor() as cursor:
            sql_function = "pkg_TestRefCursors.TestReturnCursor"
            ref_cursor = cursor.callfunc(
                sql_function, oracledb.DB_TYPE_CURSOR, [2]
            )
            column_names = [col[0] for col in ref_cursor.description]
            ref_cursor.rowfactory = lambda *row: dict(zip(column_names, row))
            expected_value = [
                {"INTCOL": 1, "STRINGCOL": "String 1"},
                {"INTCOL": 2, "STRINGCOL": "String 2"},
            ]
            self.assertEqual(ref_cursor.fetchall(), expected_value)

    def test_3927(self):
        "3927 - test using a subclassed string as bind parameter keys"

        class my_str(str):
            pass

        self.cursor.execute("truncate table TestTempTable")
        keys = {my_str("str_val"): oracledb.DB_TYPE_VARCHAR}
        self.cursor.setinputsizes(**keys)
        values = {
            my_str("int_val"): 3927,
            my_str("str_val"): "3927 - String Value",
        }
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            """,
            values,
        )
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(
            self.cursor.fetchall(), [(3927, "3927 - String Value")]
        )

    def test_3928(self):
        "3928 - test using a sequence of parameters other than a list or tuple"

        class MySeq(collections.abc.Sequence):
            def __init__(self, *data):
                self.data = data

            def __len__(self):
                return len(self.data)

            def __getitem__(self, index):
                return self.data[index]

        values_to_insert = [MySeq(1, "String 1"), MySeq(2, "String 2")]
        expected_data = [tuple(value) for value in values_to_insert]
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            """,
            values_to_insert,
        )
        self.cursor.execute(
            """
            select IntCol, StringCol1
            from TestTempTable
            order by IntCol
            """
        )
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_3929(self):
        "3929 - test an output type handler with prefetch > arraysize"

        def type_handler(cursor, metadata):
            return cursor.var(metadata.type_code, arraysize=cursor.arraysize)

        self.cursor.arraysize = 2
        self.cursor.prefetchrows = 3
        self.cursor.outputtypehandler = type_handler
        self.cursor.execute("select level from dual connect by level <= 5")
        self.assertEqual(
            self.cursor.fetchall(), [(1,), (2,), (3,), (4,), (5,)]
        )

    def test_3930(self):
        "3930 - test setinputsizes() but without binding"
        self.cursor.setinputsizes(None, int)
        sql = "select :1, : 2 from dual"
        with self.assertRaisesFullCode("ORA-01008", "DPY-4010"):
            self.cursor.execute(sql, [])

    def test_3931(self):
        "3931 - test getting FetchInfo attributes"
        type_obj = self.conn.gettype("UDT_OBJECT")
        varchar_ratio, _ = test_env.get_charset_ratios()
        test_values = [
            (
                "select IntCol from TestObjects",
                10,
                None,
                False,
                "INTCOL",
                False,
                9,
                0,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_NUMBER,
            ),
            (
                "select ObjectCol from TestObjects",
                None,
                None,
                False,
                "OBJECTCOL",
                True,
                None,
                None,
                type_obj,
                oracledb.DB_TYPE_OBJECT,
            ),
            (
                "select JsonVarchar from TestJsonCols",
                4000,
                4000 * varchar_ratio,
                True,
                "JSONVARCHAR",
                False,
                None,
                None,
                oracledb.DB_TYPE_VARCHAR,
                oracledb.DB_TYPE_VARCHAR,
            ),
            (
                "select FLOATCOL from TestNumbers",
                127,
                None,
                False,
                "FLOATCOL",
                False,
                126,
                -127,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_NUMBER,
            ),
        ]
        for (
            sql,
            display_size,
            internal_size,
            is_json,
            name,
            null_ok,
            precision,
            scale,
            typ,
            type_code,
        ) in test_values:
            self.cursor.execute(sql)
            (fetch_info,) = self.cursor.description
            self.assertIsInstance(fetch_info, oracledb.FetchInfo)
            self.assertEqual(fetch_info.display_size, display_size)
            self.assertEqual(fetch_info.internal_size, internal_size)
            if test_env.get_server_version() > (12, 2):
                self.assertEqual(fetch_info.is_json, is_json)
            self.assertEqual(fetch_info.name, name)
            self.assertEqual(fetch_info.null_ok, null_ok)
            self.assertEqual(fetch_info.precision, precision)
            self.assertEqual(fetch_info.scale, scale)
            self.assertEqual(fetch_info.type, typ)
            self.assertEqual(fetch_info.type_code, type_code)
            self.assertIsNone(fetch_info.vector_dimensions)
            self.assertIsNone(fetch_info.vector_format)

    def test_3932(self):
        "3932 - test FetchInfo repr() and str()"
        self.cursor.execute("select IntCol from TestObjects")
        (fetch_info,) = self.cursor.description
        self.assertEqual(
            str(fetch_info),
            "('INTCOL', <DbType DB_TYPE_NUMBER>, 10, None, 9, 0, False)",
        )
        self.assertEqual(
            repr(fetch_info),
            "('INTCOL', <DbType DB_TYPE_NUMBER>, 10, None, 9, 0, False)",
        )

    def test_3933(self):
        "3933 - test slicing FetchInfo"
        self.cursor.execute("select IntCol from TestObjects")
        (fetch_info,) = self.cursor.description
        self.assertEqual(fetch_info[1:3], (oracledb.DB_TYPE_NUMBER, 10))

    def test_3934(self):
        "3934 - test rowcount is zero for PL/SQL"
        self.cursor.execute("begin null; end;")
        self.assertEqual(self.cursor.rowcount, 0)
        self.cursor.execute("select user from dual")
        self.cursor.fetchall()
        self.assertEqual(self.cursor.rowcount, 1)
        self.cursor.execute("begin null; end;")
        self.assertEqual(self.cursor.rowcount, 0)

    def test_3935(self):
        "3935 - test raising no_data_found in PL/SQL"
        with self.assertRaisesFullCode("ORA-01403"):
            self.cursor.execute("begin raise no_data_found; end;")


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E2500 - Module for testing caching of configurations from config stores. No
special setup is required but the tests here will only be run if the
run_long_tests value is enabled.
"""

import time
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_extended_config_bool("run_long_tests"),
    "extended configuration run_long_tests is disabled",
)
class TestCase(test_env.BaseTestCase):
    def test_ext_2500(self):
        "E2500 - test config is cached"
        sdu = 4096
        protocol = "proto-test"
        connect_string = f"{protocol}://test_ext_2500"
        config = dict(
            connect_descriptor=test_env.get_connect_string(),
            pyo=dict(sdu=sdu),
        )

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            passed_params.set_from_config(config)
            config["pyo"]["sdu"] *= 2

        oracledb.register_protocol(protocol, hook)
        try:
            for i in range(2):
                params = oracledb.ConnectParams()
                params.parse_connect_string(connect_string)
                self.assertEqual(params.sdu, sdu)
        finally:
            oracledb.register_protocol(protocol, None)

    def test_ext_2501(self):
        "E2501 - test config cache is disabled with config_time_to_live = 0"
        sdu = 4096
        protocol = "proto-test"
        connect_string = f"{protocol}://test_ext_2501"
        config = dict(
            connect_descriptor=test_env.get_connect_string(),
            config_time_to_live=0,
            pyo=dict(sdu=sdu),
        )

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            passed_params.set_from_config(config)
            config["pyo"]["sdu"] *= 2

        oracledb.register_protocol(protocol, hook)
        try:
            for i in range(2):
                params = oracledb.ConnectParams()
                params.parse_connect_string(connect_string)
                self.assertEqual(params.sdu, sdu + i * sdu)
        finally:
            oracledb.register_protocol(protocol, None)

    def test_ext_2502(self):
        "E2502 - test config cache expiry time"
        sdu = 4096
        protocol = "proto-test"
        connect_string = f"{protocol}://test_ext_2502"
        config = dict(
            connect_descriptor=test_env.get_connect_string(),
            config_time_to_live=2,
            pyo=dict(sdu=sdu),
        )

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            passed_params.set_from_config(config)
            config["pyo"]["sdu"] *= 2

        oracledb.register_protocol(protocol, hook)
        try:
            expected_sdu = sdu
            for i in range(7):
                if i == 3 or i == 6:
                    expected_sdu *= 2
                params = oracledb.ConnectParams()
                params.parse_connect_string(connect_string)
                self.assertEqual(params.sdu, expected_sdu)
                time.sleep(0.75)
        finally:
            oracledb.register_protocol(protocol, None)

    def test_ext_2503(self):
        "E2503 - test config cache soft/hard expiry time"
        sdu = 4096
        protocol = "proto-test"
        connect_string = f"{protocol}://test_ext_2503"
        config = dict(
            connect_descriptor=test_env.get_connect_string(),
            config_time_to_live=2,
            config_time_to_live_grace_period=3,
            pyo=dict(sdu=sdu),
        )

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            if config["pyo"]["sdu"] > sdu:
                raise Exception("Arbitrary exception!")
            passed_params.set_from_config(config)
            config["pyo"]["sdu"] *= 2

        oracledb.register_protocol(protocol, hook)
        try:
            for i in range(2):
                params = oracledb.ConnectParams()
                params.parse_connect_string(connect_string)
                self.assertEqual(params.sdu, sdu)
                time.sleep(3)
            with self.assertRaisesFullCode("DPY-2056"):
                params = oracledb.ConnectParams()
                params.parse_connect_string(connect_string)
        finally:
            oracledb.register_protocol(protocol, None)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6600 - Module for testing oracledb.defaults
"""

import decimal
import os
import tempfile

import oracledb
import test_env
import unittest


class TestCase(test_env.BaseTestCase):
    def __verify_network_name_attr(self, name):
        """
        Verify that a default attribute is handled properly in both valid
        and invalid cases.
        """
        for value, ok in [
            ("valid_value", True),
            ("'contains_quotes'", False),
            ('"contains_double_quotes"', False),
            ("contains_opening_paren (", False),
            ("contains_closing_paren )", False),
            ("contains_equals =", False),
            ("contains_trailing_slash\\", False),
        ]:
            args = {}
            args[name] = value
            if ok:
                with test_env.DefaultsContextManager(name, value):
                    cp = oracledb.ConnectParams(**args)
                    self.assertEqual(getattr(cp, name), value)
            else:
                with self.assertRaisesFullCode("DPY-3029"):
                    with test_env.DefaultsContextManager(name, value):
                        pass

    def test_6600(self):
        "6600 - test setting defaults.arraysize"
        with test_env.DefaultsContextManager("arraysize", 50):
            conn = test_env.get_connection()
            cursor = conn.cursor()
            self.assertEqual(cursor.arraysize, oracledb.defaults.arraysize)

    def test_6601(self):
        "6601 - test getting decimals with defaults.fetch_decimals=True"
        with test_env.DefaultsContextManager("fetch_decimals", True):
            self.cursor.execute("select 9 from dual")
            (result,) = self.cursor.fetchone()
            self.assertIsInstance(result, decimal.Decimal)

    def test_6602(self):
        "6602 - test getting string lob with defaults.fetch_lobs=False"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute("select to_clob('Hello world') from dual")
            (result,) = self.cursor.fetchone()
            self.assertIsInstance(result, str)

    def test_6603(self):
        "6603 - test setting defaults.prefetchrows"
        with test_env.DefaultsContextManager("prefetchrows", 20):
            conn = test_env.get_connection()
            cursor = conn.cursor()
            self.assertEqual(
                cursor.prefetchrows, oracledb.defaults.prefetchrows
            )

    def test_6604(self):
        "6604 - test setting defaults.stmtcachesize (pool)"
        new_stmtcachesize = 15
        with test_env.DefaultsContextManager("stmtcachesize", 40):
            pool = test_env.get_pool()
            self.assertEqual(
                pool.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            conn = pool.acquire()
            self.assertEqual(
                conn.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            pool = test_env.get_pool(stmtcachesize=new_stmtcachesize)
            self.assertEqual(pool.stmtcachesize, new_stmtcachesize)
            conn = pool.acquire()
            self.assertEqual(conn.stmtcachesize, new_stmtcachesize)

    def test_6605(self):
        "6605 - test setting defaults.stmtcachesize (standalone connection)"
        new_stmtcachesize = 25
        with test_env.DefaultsContextManager("stmtcachesize", 50):
            conn = test_env.get_connection()
            self.assertEqual(
                conn.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            conn = test_env.get_connection(stmtcachesize=new_stmtcachesize)
            self.assertEqual(conn.stmtcachesize, new_stmtcachesize)

    def test_6606(self):
        "6606 - fetch_lobs does not affect LOBS returned as OUT binds"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            var = self.cursor.var(oracledb.DB_TYPE_CLOB)
            self.cursor.execute(
                "begin :value := to_clob('test clob'); end;",
                value=var,
            )
            self.assertIsInstance(var.getvalue(), oracledb.LOB)

    def test_6607(self):
        "6607 - test setting defaults.config_dir"
        with tempfile.TemporaryDirectory() as temp_dir:
            new_temp_dir = os.path.join(temp_dir, "subdir")
            with test_env.DefaultsContextManager("config_dir", temp_dir):
                self.assertEqual(oracledb.defaults.config_dir, temp_dir)
                params = oracledb.ConnectParams()
                self.assertEqual(params.config_dir, temp_dir)
                params = oracledb.ConnectParams(config_dir=new_temp_dir)
                self.assertEqual(params.config_dir, new_temp_dir)

    def test_6608(self):
        "6608 - test setting defaults.stmtcachesize (ConnectParams)"
        new_stmtcachesize = 35
        with test_env.DefaultsContextManager("stmtcachesize", 60):
            params = oracledb.ConnectParams()
            self.assertEqual(
                params.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            params = oracledb.ConnectParams(stmtcachesize=new_stmtcachesize)
            self.assertEqual(params.stmtcachesize, new_stmtcachesize)

    def test_6609(self):
        "6609 - test defaults.stmtcachesize persists after setting it again"
        value = 50
        new_value = 29
        with test_env.DefaultsContextManager("stmtcachesize", value):
            pool = test_env.get_pool()
            pooled_conn = pool.acquire()
            params = oracledb.ConnectParams()
            standalone_conn = test_env.get_connection()
            with test_env.DefaultsContextManager("stmtcachesize", new_value):
                self.assertEqual(pool.stmtcachesize, value)
                self.assertEqual(pooled_conn.stmtcachesize, value)
                self.assertEqual(params.stmtcachesize, value)
                self.assertEqual(standalone_conn.stmtcachesize, value)
                pool = test_env.get_pool()
                pooled_conn = pool.acquire()
                params = oracledb.ConnectParams()
                standalone_conn = test_env.get_connection()
                self.assertEqual(pool.stmtcachesize, new_value)
                self.assertEqual(pooled_conn.stmtcachesize, new_value)
                self.assertEqual(params.stmtcachesize, new_value)
                self.assertEqual(standalone_conn.stmtcachesize, new_value)

    def test_6610(self):
        "6610 - test setting defaults.terminal"
        with test_env.DefaultsContextManager("terminal", "newterminal"):
            params = oracledb.ConnectParams()
            self.assertEqual(params.terminal, oracledb.defaults.terminal)

    def test_6611(self):
        "6611 - test setting defaults.driver_name"
        with test_env.DefaultsContextManager("driver_name", "newdriver"):
            params = oracledb.ConnectParams()
            self.assertEqual(params.driver_name, oracledb.defaults.driver_name)

    def test_6612(self):
        "6612 - test setting defaults.program attribute"
        self.__verify_network_name_attr("program")

    def test_6613(self):
        "6613 - test setting defaults.machine attribute"
        self.__verify_network_name_attr("machine")

    def test_6614(self):
        "6614 - test setting defaults.osuser attribute"
        self.__verify_network_name_attr("osuser")

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support program yet",
    )
    def test_6615(self):
        "6615 - test program with two pools"
        default_value = "defaultprogram"
        new_value = "newprogram"
        verify_sql = (
            "select program from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        with test_env.DefaultsContextManager("program", default_value):

            # create pool using default value
            pool = test_env.get_pool()
            with pool.acquire() as conn:
                with conn.cursor() as cursor:
                    cursor.execute(verify_sql)
                    (fetched_value,) = cursor.fetchone()
                    self.assertEqual(fetched_value, default_value)
            pool.close()

            # create pool using new value
            pool = test_env.get_pool(program=new_value)
            with pool.acquire() as conn:
                with conn.cursor() as cursor:
                    cursor.execute(verify_sql)
                    (fetched_value,) = cursor.fetchone()
                    self.assertEqual(fetched_value, new_value)
            pool.close()


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# encoders.pyx
#
# Cython file defining the low-level encoding routines used by the driver.
#------------------------------------------------------------------------------

cdef inline void encode_uint16be(char_type *buf, uint16_t value):
    """
    Encodes a 16-bit integer in big endian order (most significant byte first).
    """
    buf[0] = <char_type> ((value >> 8) & 0xff)
    buf[1] = <char_type> (value & 0xff)


cdef inline void encode_uint16le(char_type *buf, uint16_t value):
    """
    Encodes a 16-bit integer in big endian order (most significant byte first).
    """
    buf[1] = <char_type> ((value >> 8) & 0xff)
    buf[0] = <char_type> (value & 0xff)


cdef inline void encode_uint32be(char_type *buf, uint32_t value):
    """
    Encodes a 32-bit integer in big endian order (most significant byte first).
    """
    buf[0] = <char_type> ((value >> 24) & 0xff)
    buf[1] = <char_type> ((value >> 16) & 0xff)
    buf[2] = <char_type> ((value >> 8) & 0xff)
    buf[3] = <char_type> (value & 0xff)


cdef inline void encode_uint64be(char_type *buf, uint64_t value):
    """
    Decodes a 64-bit integer in big endian order (most significant byte first).
    """
    buf[0] = <char_type> ((value >> 56) & 0xff)
    buf[1] = <char_type> ((value >> 48) & 0xff)
    buf[2] = <char_type> ((value >> 40) & 0xff)
    buf[3] = <char_type> ((value >> 32) & 0xff)
    buf[4] = <char_type> ((value >> 24) & 0xff)
    buf[5] = <char_type> ((value >> 16) & 0xff)
    buf[6] = <char_type> ((value >> 8) & 0xff)
    buf[7] = <char_type> (value & 0xff)
